%% T-61.2010 Datasta tietoon / T#3 / 4
% Jukka Parviainen, 21.11.2006

clear;         % poista muistista muuttujat
clc;           % puhdista komentoikkunan ruutu
close all;     % sulje kaikki ikkunat

%% c-means: kiisseli
% Also known as (more commonly?) "k-means".
%
% On a certain basic data analysis course of TKK the students were requested
% to pronounce word "kiisseli" and record the sound in WAV format
% see e.g.
%|http://www.cis.hut.fi/Opinnot/T-61.2010/Laskarit/Kiisseli/|
%
% Ten samples of these were picked that were concatenated and 
% written as a single audio file.
% Each sample was rescaled so that the maximum values were set to be one.

% Mel-frequency cepstral coefficients (MFCC)were calculated for 
% windows of length 40 ms that started with 20 ms intervals (20ms overlap).
% The MFCC are given by
%
% MFCC = ifft(log10(abs(fft(x))^2)
%
% The MFCC's are very often used as features 
% in speech recognition algorithms. They have been demonstrated
% to work well and remove correlations.
% This time the coefficients were calculated for 13 frequency
% bands (0-11025 Hz); On Mel scale the bands are narrower on lower
% frequencies
%
% The original sound recording has 22050 samples a 8 bits per
% second, that is 22050 bytes. 13 coefficients are calculated every
% 20ms, that is 50*13=650 per second (each coefficient is represented with 
% ~4 bytes)
%
% The original signal cannot be reconstructed from real cepstral coefficients
% but they are intended for analysing the signal.
%
% In the following we are going to cluster the coefficients
% with the c-means algorithm and listen to the sound samples 
% corresponding to the clusters
%
% We also calculate the energy of 13 evenly spaced frequency bands
% directly from the magnitude of Fourier-transform.
% Let's denote these coefficients "short-time Fourier-transform 
% coefficients" (STFTC).
% We perform the c-means clustering also to these. 

%% Load and listen to the sound file.
[X, Fs] = wavread('kiisseli_10x.wav');
soundsc(X, Fs);

%% Plot the sound signal
% Upper plot: the waveform, the time in seconds on x-axis
% lower plot: the spectrogram, x-axis: time in seconds, y-axis: frequency


t = [1 : length(X)]/Fs;

figure(41); clf;
subplot(2,1,1);
plot(t, X);
set(gca,'XLim',[t(1) t(end)]);
title('Kiisseli x 10');
grid on;
subplot(2,1,2);
specgram(X, [], Fs);
% In newer versions of Matlab:
%spectrogram(X, [], [], [], Fs, 'yaxis');

%% Load the readily calculated MFCC coefficients
% The Mel-cepstrum is calculated with the
% Matlab toolbax available in the net,
% |http://labrosa.ee.columbia.edu/matlab/rastamat/|
% "PLP and RASTA (and MFCC, and inversion) in Matlab".
% The command used:
% [mfcc] = melfcc(X, Fs, 'maxfreq', Fs/2, 'numcep', 13, 'nbands', 14, 'fbtype', 'fcmel', 'dcttype', 1, 'usecmp', 1, 'wintime', 0.04, 'hoptime', 0.02, 'preemph', 0, 'dither', 1);
%
%  Thus a 13-dimensional MFCC every 20ms is obtained from the sound. 


load dt_t3_m4_mfcc.mat
[a, b] = size(mfcc);
disp(['Read ' num2str(b) ' ' num2str(a) '-D MFCC-coefficients']);

%% Calculate also the spectral coefficients
% Short-time Fourier-transform coefficients a la Jukka:
% Compute spectra first and combine to 13 equal-wide bands.
winlen = round(0.04*Fs);
noverlap = round(0.02*Fs);
nfft = 24;
[XTF] = abs(specgram(X, 1300, Fs, winlen, noverlap));
for k = [1 : 13]
    stftc(k, :) = sum(XTF((k-1)*50+1: k*50, :));
end;
% Newest version of Matlab:
%[stftc] = abs(spectrogram(X, winlen, noverlap, nfft));
disp(['Calculated ' num2str(size(stftc, 2)) ' ' num2str(size(stftc,1)) '-D STFTC-coefficients']);

%% Choose the number of clusters in c-means clustering
% Specify the number of clusters in variable c.
% How do you know the correct number of clusters?

c = 3;
disp(['Number of clusters ' num2str(c) ' chosen.']);

%% Plot the MFCC and STFTC coefficients

figure(42); clf;
subplot(3,1,1);
plot(t, X);
grid on;
set(gca,'XLim', [t(1) t(end)]);
title('Waveform: 10 x kiisseli');
subplot(3,1,2);
imagesc(flipud((mfcc)));
title(['The ' num2str(length(mfcc)) ' MFCC coefficients calculated from the signal']);
subplot(3,1,3);
imagesc(flipud((stftc)));
title(['The ' num2str(length(stftc)) ' STFTC coefficients calculated from the signal']);

%% Cluster the MFCC
% Use the command kmeans, "help kmeans".
% The desired number of clusters is c.

disp('Clustering the  MFCC coefficients:');
idx_mfcc = kmeans(mfcc', c, 'display','iter')';

subplot(3,1,2);
hold on;
pl1 = plot(idx_mfcc);
set(pl1, 'LineWidth', 3, 'Color', [0 0 0]);

%% Cluster the Fourier coefficients, i.e. STFTC
% The desired number of clusters is c.

disp('Klustering the STFTC coefficients:');
idx_stftc = kmeans(stftc', c, 'display','iter')';

subplot(3,1,3);
hold on;
pl2 = plot(idx_stftc);
set(pl2, 'LineWidth', 3, 'Color', [0 0 0]);

%% Gather the results in sound vectors
% The resulting vectors are approximately doubled in length.

X_mfcc = cell(c, 1);
X_stftc = cell(c, 1);

for s = [1 : c]
    dummy = [];
    for m = [find(idx_mfcc == s)]
        dummy = [dummy; X(round((m-1)*441)+1 : round((m+1)*441)-1)];
    end;
    X_mfcc{s} = dummy;
end;

for s = [1 : c]
    dummy = [];
    for m = [find(idx_stftc == s)]
        dummy = [dummy; X(round((m-1)*441)+1 : round((m+1)*441)-1)];
    end;
    X_stftc{s} = dummy;
end;

%% Listen to the results of the clustering
% Instead if visualising we listen to the clusters
%
% How would you characterise the resulting clusters?

disp('Listen to the MFCC-clusters:');
for s = [1 : c]
    disp(['Cluster #' num2str(s) '/' num2str(c) ' (MFCC): ' num2str(length(X_mfcc{s})/Fs) 's']);
    soundsc(X_mfcc{s}, Fs);
    disp('Hit any key to continue...');
    pause;
end;

disp('Listen to the STFTC-clusters:');
for s = [1 : c]
    disp(['Cluster #' num2str(s) '/' num2str(c) ' (STFTC): ' num2str(length(X_stftc{s})/Fs) 's']);
    soundsc(X_stftc{s}, Fs);
    disp('Hit any key to continue...');
    pause;
end;

%% Observations
% Note that the initial cluster centers are chosen randomly in c-means
% clustering. This may affect the result of the clustering significantly.
% Therefore run several times with the same value of c.

